;; File:
;;
;;   enhanced-buffer-rings.el
;;
;; Version:
;;
;;   v1.0
;;
;; Author(s):
;;
;;   Charles K. Hines
;;
;; Licensing:
;;
;;   GPL v2 or later (see http://www.gnu.org/licenses/gpl.html for details)
;;
;; Description:
;;
;;   Elisp to assign a buffer to a specific buffer ring and functions to walk
;;   through those rings.
;;
;; Usage:
;;
;;   (require 'enhanced-buffer-rings)
;;   ;; or
;;   (autoload 'ebr::add-buffer-to-ring "enhanced-buffer-rings" nil t)
;;   (autoload 'ebr::remove-buffer-from-ring "enhanced-buffer-rings" nil t)
;;   (autoload 'ebr::previous-buffer-in-ring "enhanced-buffer-rings" nil t)
;;   (autoload 'ebr::next-buffer-in-ring "enhanced-buffer-rings" nil t)
;;   (autoload 'ebr::previous-buffer-in-global-ring "enhanced-buffer-rings" nil t)
;;   (autoload 'ebr::next-buffer-in-global-ring "enhanced-buffer-rings" nil t)
;;
;;   (global-set-key [(f11)] 'ebr::add-buffer-to-ring)
;;   (global-set-key [(f12)] 'ebr::remove-buffer-from-ring)
;;   (global-set-key [(meta prior)] 'ebr::previous-buffer-in-ring)
;;   (global-set-key [(meta next)] 'ebr::next-buffer-in-ring)
;;   (global-set-key [(alt prior)] 'ebr::previous-buffer-in-global-ring)
;;   (global-set-key [(alt next)] 'ebr::next-buffer-in-global-ring)
;;
;; Notes:
;;
;;   I wrote this to replace these two somewhat deficient functions:
;;
;;     (defun buffer-ring ()
;;       "Works around a ring of buffers.
;;        Steps to the next buffer in list, putting the current buffer
;;        at the end of the selection list."
;;       (interactive)
;;       (bury-buffer (current-buffer))
;;       (switch-to-buffer (other-buffer)))
;;     ;; borrowed code from mouse-unbury-buffer to create the reverse function:
;;     (defun buffer-ring-backwards ()
;;       "Works backwards throught a ring of buffers"
;;       (interactive)
;;       (let* ((bufs (buffer-list))
;;              (entry (1- (length bufs)))
;;              val)
;;         (while (not (setq val (nth entry bufs)
;;                           val (and (/= (aref (buffer-name val) 0)
;;                                        ? )
;;                                    val)))
;;           (setq entry (1- entry)))
;;         (switch-to-buffer val)))
;;
;;   that I used to have bound to M-PageUp and M-PageDown (respectively).  I
;;   got really annoyed at the behavior of those functions when I opened up a
;;   new file in another window because it "messed up" my desired ordering by
;;   placing the new buffer as next in the list, which of course makes sense
;;   but was still annoying when it "messed" with the order in another window.
;;
;;   So I came up with the idea of having multiple buffer rings so I could add
;;   all "related" buffers to a single ring and ONLY get those buffers when
;;   rotating through them.  So I could do things like have all the header
;;   files for one part of the project I was working on in one ring, all the
;;   source related to those in another, the headers for a second (unrelated)
;;   part of the project in a third ring, the source for those in a fourth,
;;   etc.
;;
;;   Also, the old functions I'm replacing are preserved here as
;;   {next,previous}-buffer-in-global-ring since they are useful at times as
;;   well.
;;
;;   The current implementation is a bit simplistic: a hardcoded list of 10
;;   rings that can hold a default of 20 buffers each.  Using a prefix value
;;   to the add-buffer-to-ring function controls which ring the buffer gets
;;   added to (defaults to 1).
;;
;;   My original idea involved "named" rings, giving an arbitrary number of
;;   them essentially with a description of the types of buffers on the ring
;;   if you felt like being that creative with the ring names, and I may
;;   enhance this code to operate that way someday but I think this
;;   implementation will be good enough for now.
;;
;;   I might also code up some new ring code in the future to allow for
;;   resizing the rings dynamically instead of having to pick some arbitary
;;   number of buffers allowed per ring.
;;
;;   Might be interesting to add some code here in the future to automatically
;;   add buffers to rings as files are loaded (or the buffers created), based
;;   on mode, extension, directory, etc.
;;
;; History:
;;   15 Feb 2003 (CKH) Created v1.0
;;

(require 'ring)
(require 'cl) ;; incf

;; is there a function to create a vector w/ nil values for each entry
;; specifying the size of the vector so I could make the number of rings
;; a variable parameter, or would I have to switch to this being a list
;; instead of a vector to do that (or rather, make it a list so I don't have
;; to have a set number of rings, grow the list as necessary)?
(defvar ebr::buffer-ring-list [nil nil nil nil nil nil nil nil nil nil]
  "global array (vector) of buffer rings (10 rings total)"
  )

;; ugh, wish rings were a little more open ended...perhaps I should write a
;; new ring package implementing growable/shrinkable rings?
(defvar ebr::buffer-ring-list-max-buffers-per-ring 20
  "max number of buffers that can be contained per ring")

(defun ebr::add-buffer-to-ring (&optional rname buf)
  "Adds buffer (default current) to a specific buffer ring (default 1 if no prefix arg specified)."
  (interactive "p")
  ;; rname has to be >= 0 and <= 9 (is there an easier way to check this?  or
  ;; can/should I just let the array access funcs deal with it?)
  (if (not (and (>= rname 0)
                (<= rname 9)))
      (error 'args-out-of-range
             "prefix arg (ring number) must be in the range 0-9 (inclusive)"))
  ;; create new ring if need be
  (if (null (aref ebr::buffer-ring-list rname))
      (aset ebr::buffer-ring-list rname
            (make-ring ebr::buffer-ring-list-max-buffers-per-ring)))
  ;; should I remove this buffer from any other rings it's in at this point?
  ;; probably:
  (ebr::remove-buffer-from-ring buf)
  ;; now add buffer to ring (modifying order if alread in there)
  (if (null buf)
      (setq buf (current-buffer)))
  (ebr::ring+::ring-insert-new (aref ebr::buffer-ring-list rname)
                               buf)
  (message "buffer %s added to ring %d" buf rname)
  )

(defun ebr::remove-buffer-from-ring (&optional buf)
  "Removes buffer (default current) from the ring it's in"
  (interactive)
  (let ((rname)
        (rng))
    (if (null buf)
        (setq buf (current-buffer)))
    ;; just in case it accidentally ended up in multiple rings (shouldn't
    ;; normally occur):
    (while (setq rname (ebr::get-ring-for-buffer buf))
      ;; remove buffer from ring
      (setq rng (aref ebr::buffer-ring-list rname))
      (ring-remove rng (ebr::ring+::ring-member buf rng))
      (message "buffer %s removed from ring %d" buf rname)
      )))

;; hm, currently returning index of ring into array, perhaps I should simply
;; return the ring itself?  Maybe split into 2 functions since both would be
;; handy.
(defun ebr::get-ring-for-buffer (&optional buf)
  "returns ring buffer buf (default current) appears in"
  (interactive) ;; might be handy to call via M-x to see if buf is in a ring
		;; already or not
  (let ((rname nil)
        (cnt 0)
        (idx)
        (rng)
        (rname nil))
    (if (null buf)
        (setq buf (current-buffer)))
    (while (and (<= cnt 9) (null rname))
      (setq rng (aref ebr::buffer-ring-list cnt))
      (if rng
          (if (setq idx (ebr::ring+::ring-member buf rng))
              (setq rname cnt)))
      (incf cnt))
    rname))

(defun ebr::next-buffer-in-ring (&optional buf)
  "Switches to next buffer a specific named buffer ring (based on ring
buffer buf (default current) appears in).  If the buffer isn't in any
ring, we'll pretend that buffer-list is a ring and go to the next one
there."
  (interactive)
  (if (null buf)
      (setq buf (current-buffer)))
  (let* ((rname (ebr::get-ring-for-buffer buf))
         (rng (if rname (aref ebr::buffer-ring-list rname) nil)))
    (if rng
        (switch-to-buffer
         (ring-ref rng (1+ (ebr::ring+::ring-member buf rng))))
      ;;(message "not in any ring?")
      ;; use "global" buffer ring
      (ebr::next-buffer-in-global-ring))))

(defun ebr::previous-buffer-in-ring (&optional buf)
  "Switches to previous buffer a specific named buffer ring (based on ring
current buffer appears in).  If the buffer isn't in any ring, we'll pretend
that buffer-list is a ring and go to the previous one there."
  (interactive)
  (if (null buf)
      (setq buf (current-buffer)))
  (let* ((rname (ebr::get-ring-for-buffer buf))
         (rng (if rname (aref ebr::buffer-ring-list rname) nil)))
    (if rng
        (switch-to-buffer
         (ring-ref rng (1- (ebr::ring+::ring-member buf rng))))
      ;;(message "not in any ring?")
      ;; use "global" buffer ring
      (ebr::previous-buffer-in-global-ring))))

(defun ebr::previous-buffer-in-global-ring ()
  "Works like ebr::previous-buffer-in-ring but operates on the \"global\"
list (buffer-list)"
  (interactive)
  (bury-buffer (current-buffer))
  (switch-to-buffer (other-buffer)))

(defun ebr::next-buffer-in-global-ring ()
  "Works like ebr::next-buffer-in-ring but operates on the \"global\"
list (buffer-list)"
  (interactive)
  (let* ((bufs (buffer-list))
         (entry (1- (length bufs)))
         val)
    (while (not (setq val (nth entry bufs)
                      val (and (/= (aref (buffer-name val) 0)
                                   ? )
                               val)))
      (setq entry (1- entry)))
    (switch-to-buffer val)))

;; borrowed these two functions (ring-member and ring-insert-new) from
;; ring+.el written by Drew Adams and found on emacswiki.org so I wouldn't
;; have to require ring+

(defun ebr::ring+::ring-member (item ring)
  "Returns index of ITEM if on RING, else nil.  Comparison via `equal'.
The index is 0-based."
  (let ((ind 0)
        (len (1- (ring-length ring)))
        (memberp nil))
    (while (and (<= ind len)
                (not (setq memberp (equal item (ring-ref ring ind)))))
      (incf ind))
    (and memberp ind)))

(defun ebr::ring+::ring-insert-new (ring item)
  "`ring-remove' ITEM from RING, then `ring-insert' it back onto RING.
This ensures that there is only one ITEM on RING.
Returns the new RING."
  (let (ind)
    (while (setq ind (ebr::ring+::ring-member item ring))
      (ring-remove ring ind)))
  (ring-insert ring item)
  ring)




(provide 'enhanced-buffer-rings)

